// Copyright 2022 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package xflags

import (
	"flag"
	"fmt"
	"github.com/stretchr/testify/assert"
	"reflect"
	"strings"
	"testing"
)

func (s *Settings[C]) Execute() *Settings[C] {
	if len(s.errors) > 0 {
		return s
	}

	if s.command == nil {
		s.errors = append(s.errors, MissingCommandError)
		return s
	}

	if !s.command.flagSet.Parsed() {
		s.errors = append(s.errors, NotParsedError)
	} else {
		s.wasExecuted = callCmdFunctions(s, s.command.commands)
	}

	return s
}

func callCmdFunctions[C any](settings *Settings[C], commands []*cmd[C]) bool {

	wasExecuted := false
	shouldExecute := false

	var availableCommands []string
	currentCommand := ""

	for _, command := range commands {
		if command.flagSet.Parsed() {

			shouldExecute = true
			currentCommand = command.name

			if len(command.commands) > 0 {
				r := callCmdFunctions(settings, command.commands)
				if r {
					wasExecuted = true
				}
			}

			if !wasExecuted {
				f := reflect.ValueOf(&command.settings.definitions).MethodByName(command.functionName)
				if f.IsValid() {
					m := command.settings
					in := []reflect.Value{reflect.ValueOf(m)}
					f.Call(in)
					wasExecuted = true
				}
			}

			break
		} else {
			availableCommands = append(availableCommands, command.name)
		}

	}

	if shouldExecute {
		if !wasExecuted {
			settings.errors = append(settings.errors, newMissingFunctionError(currentCommand))
			return false
		}

		return true
	}

	if len(availableCommands) > 0 {
		if settings.hint == "" {
			settings.hint = fmt.Sprintf("Did you mean: %v?", strings.Join(availableCommands, ", "))
		}
	}

	settings.errors = append(settings.errors, MissingCommandError)

	return false

}

// HelpRequested indicates if the help flag was set.
func (s *Settings[C]) HelpRequested() bool {

	for _, err := range s.errors {
		if err == flag.ErrHelp {
			return true
		}
	}

	return false
}

// MissingCommandError is returned if no command was found.
func (s *Settings[C]) MissingCommand() bool {

	for _, err := range s.errors {
		if err == MissingCommandError {
			return true
		}
	}

	return false
}

// WasExecuted returns true if the call function was executed
func (s *Settings[C]) WasExecuted() bool {
	return s.wasExecuted
}

func TestWrongDefinitionType(t *testing.T) {
	c := New("root", 2)
	c.Parse([]string{"test"})
	c.Execute()
	assert.True(t, c.HasErrors())
}

type testExecuteCommandStruct struct {
	Command1 struct {
	} `command:"command1" description:"Command 1" callback:"command1Callback" `
	Command2 struct {
		Command3 struct {
		} `command:"command3" description:"Command 3" callback:"command3Callback" call:"DoCmd3"`
	} `command:"command2" description:"Command 2" callback:"command2Callback" `
}

func (c *testExecuteCommandStruct) DoCmd3(s *Settings[testExecuteCommandStruct]) {
}

func (c *testExecuteCommandStruct) command1Callback(args []string) {
	fmt.Println("command1Callback", args)
}

func TestExecute1(t *testing.T) {
	c := New("root", testExecuteCommandStruct{})
	c.Parse([]string{"command2", "command3", "commandX"})
	c.Execute()
	assert.False(t, c.HasErrors())
}
